"""All physical constants used in other code."""
# TODO Constants could be used more
global kb, h, hbar

kb = 1.380649e-23
h = 6.62607015e-34
hbar = 1.054571817e-34

PREFIXES = {
    "K": 1e3,
    "M": 1e6,
    "G": 1e9,
    "T": 1e12,
    "m": 1e-3,
    "µ": 1e-6,
    "n": 1e-9,
    "p": 1e-12,
}
